package com.atlassian.example.jira.jqlfunc;

import com.atlassian.jira.JiraDataType;
import com.atlassian.jira.JiraDataTypes;
import com.atlassian.jira.jql.operand.QueryLiteral;
import com.atlassian.jira.jql.query.QueryCreationContext;
import com.atlassian.jira.plugin.jql.function.ClauseSanitisingJqlFunction;
import com.atlassian.jira.plugin.jql.function.JqlFunction;
import com.atlassian.jira.plugin.jql.function.JqlFunctionModuleDescriptor;
import com.atlassian.jira.project.Project;
import com.atlassian.jira.project.ProjectManager;
import com.atlassian.jira.security.PermissionManager;
import com.atlassian.jira.security.Permissions;
import com.atlassian.jira.security.roles.ProjectRole;
import com.atlassian.jira.security.roles.ProjectRoleActors;
import com.atlassian.jira.security.roles.ProjectRoleManager;
import com.atlassian.jira.util.MessageSet;
import com.atlassian.jira.util.MessageSetImpl;
import com.atlassian.query.clause.TerminalClause;
import com.atlassian.query.operand.FunctionOperand;
import com.opensymphony.user.User;
import org.apache.commons.lang.StringUtils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * A simple example of a JQL function that returns all the members of a particular project role.
 * <p/>
 * The usage of the function is {@code roleMembers(&lt;roleName&gt;,[&lt;project&gt;]*)}
 * <p/>
 * The roleName cannot be blank. The project parameter(s) can be used to limit the search for members to the passed
 * project(s). If no project is specified, then the function will search through all projects that the searching user
 * can browse.
 */
public class RoleFunction implements JqlFunction, ClauseSanitisingJqlFunction
{
    private final ProjectRoleManager roleManager;
    private final PermissionManager permissionManager;
    private final ProjectManager projectManager;

    private volatile JqlFunctionModuleDescriptor descriptor;

    public RoleFunction(final ProjectRoleManager roleManager, final PermissionManager permissionManager, final ProjectManager projectManager)
    {
        this.roleManager = roleManager;
        this.permissionManager = permissionManager;
        this.projectManager = projectManager;
    }

    public void init(final JqlFunctionModuleDescriptor descriptor)
    {
        this.descriptor = descriptor;
    }

    public MessageSet validate(final User searcher, final FunctionOperand operand, final TerminalClause terminalClause)
    {
        MessageSet messages = new MessageSetImpl();
        final List<String> arguments = operand.getArgs();

        //Make sure we have the correct number of arguments.
        if (arguments.isEmpty())
        {
            messages.addErrorMessage(descriptor.getI18nBean().getText("rolefunc.bad.num.arguments", operand.getName()));
            return messages;
        }

        //Make sure the role is valid.
        final String role = arguments.get(0);
        validateRole(role, messages);

        //Make sure the project arguments are valid if provided.
        if (arguments.size() > 1)
        {
            for (String project : arguments.subList(1, arguments.size()))
            {
                validateProject(searcher, project, messages);
            }
        }

        return messages;
    }

    public List<QueryLiteral> getValues(final QueryCreationContext queryCreationContext, final FunctionOperand operand, final TerminalClause terminalClause)
    {
        final List<String> arguments = operand.getArgs();
        //Can't do anything when no argument is specified. This is an error so return empty list.
        if (arguments.isEmpty())
        {
            return Collections.emptyList();
        }

        final ProjectRole projectRole = lookupRole(arguments.get(0));
        //Role not in system, then do nothing.
        if (projectRole == null)
        {
            return Collections.emptyList();
        }

        final Collection<User> users;
        if (arguments.size() > 1)
        {
            //Projects are specified, then look at those projects.
            users = getUsersForProjects(queryCreationContext, projectRole, arguments.subList(1, arguments.size()));
        }
        else
        {
            //No project specified, then look at all projects.
            users = getUsersForAllProjects(queryCreationContext, projectRole);
        }

        //Covert all the users to query literals.
        final List<QueryLiteral> literals = new ArrayList<QueryLiteral>(users.size());
        for (User user : users)
        {
            literals.add(new QueryLiteral(operand, user.getName()));
        }

        return literals;
    }

    public boolean isList()
    {
        return true;
    }

    public int getMinimumNumberOfExpectedArguments()
    {
        return 1;
    }

    public String getFunctionName()
    {
        return "roleMembers";
    }

    public JiraDataType getDataType()
    {
        return JiraDataTypes.USER;
    }

    public FunctionOperand sanitiseOperand(final User user, final FunctionOperand functionOperand)
    {
        final List<String> arguments = functionOperand.getArgs();

        //We only sanitise projects, so just return the original function if there are no projects. 
        if (arguments.size() <= 1)
        {
            return functionOperand;
        }

        boolean argChanged = false;
        final List<String> newArgs = new ArrayList<String>(arguments.size());
        newArgs.add(arguments.get(0));
        for (final String argument : arguments.subList(1, arguments.size()))
        {
            final Project project = findProject(argument);
            if (project != null && !permissionManager.hasPermission(Permissions.BROWSE, project, user))
            {
                newArgs.add(project.getId().toString());
                argChanged = true;
            }
            else
            {
                newArgs.add(argument);
            }
        }

        if (argChanged)
        {
            return new FunctionOperand(functionOperand.getName(), newArgs);
        }
        else
        {
            return functionOperand;
        }
    }

    private boolean validateRole(final String roleString, final MessageSet messages)
    {
        if (StringUtils.isBlank(roleString))
        {
            messages.addErrorMessage(descriptor.getI18nBean().getText("rolefunc.role.blank.name"));
            return false;
        }

        final ProjectRole projectRole = roleManager.getProjectRole(roleString);
        if (projectRole == null)
        {
            messages.addErrorMessage(descriptor.getI18nBean().getText("rolefunc.role.does.not.exist", roleString));
            return false;
        }
        return true;
    }

    private ProjectRole lookupRole(String role)
    {
        if (StringUtils.isBlank(role))
        {
            return null;
        }

        return roleManager.getProjectRole(role);
    }

    private boolean validateProject(final User searcher, final String project, final MessageSet messages)
    {
        if (StringUtils.isBlank(project))
        {
            messages.addErrorMessage(descriptor.getI18nBean().getText("rolefunc.project.blank.name"));
            return false;
        }

        final Project projectObj = findProject(project);
        if (projectObj == null || !permissionManager.hasPermission(Permissions.BROWSE, projectObj, searcher))
        {
            messages.addErrorMessage(descriptor.getI18nBean().getText("rolefunc.project.does.not.exist", project));
            return false;
        }

        return true;
    }

    private Project lookupProject(final QueryCreationContext ctx, final String project)
    {
        if (StringUtils.isBlank(project))
        {
            return null;
        }

        final Project objByName = findProject(project);
        if (objByName == null || (!ctx.isSecurityOverriden() && !permissionManager.hasPermission(Permissions.BROWSE, objByName, ctx.getUser())))
        {
            return null;
        }

        return objByName;
    }

    private Collection<Project> lookupProjects(final QueryCreationContext ctx, final Collection<String> projects)
    {
        final List<Project> returnValue = new ArrayList<Project>(projects.size());
        for (String project : projects)
        {
            Project projObj = lookupProject(ctx, project);
            if (projObj != null)
            {
                returnValue.add(projObj);
            }
        }

        return returnValue;
    }

    private Collection<User> getUsersForProjects(final QueryCreationContext ctx, final ProjectRole role, final Collection<String> projects)
    {
        final Collection<Project> projectObjs = lookupProjects(ctx, projects);
        return getUsers(role, projectObjs);
    }

    private Collection<User> getUsersForAllProjects(final QueryCreationContext ctx, final ProjectRole role)
    {
        final Collection<Project> projects;
        if (ctx.isSecurityOverriden())
        {
            projects = projectManager.getProjectObjects();
        }
        else
        {
            projects = permissionManager.getProjectObjects(Permissions.BROWSE, ctx.getUser());
        }
        return getUsers(role, projects);
    }

    private Collection<User> getUsers(final ProjectRole projectRole, final Collection<Project> projects)
    {
        final Set<User> users = new HashSet<User>();
        for (Project project : projects)
        {
            final ProjectRoleActors roleActors = roleManager.getProjectRoleActors(projectRole, project);
            if (roleActors != null)
            {
                users.addAll(roleActors.getUsers());
            }
        }
        return users;
    }

    private Project findProject(final String project)
    {
        if (StringUtils.isBlank(project))
        {
            return null;
        }

        final Project objByName = projectManager.getProjectObjByName(project);
        if (objByName == null)
        {
            try
            {
                final Long id = new Long(project);
                return projectManager.getProjectObj(id);
            }
            catch (NumberFormatException ignored)
            {
                return null;
            }
        }
        else
        {
            return objByName;
        }
    }
}
